/*
 * Harissa, copyright (C) Irisa/Inria 1996
 *
 * $Id: harissa_types.h,v 1.18 1997/12/10 17:29:37 ups Exp $
 *
 * $Log: harissa_types.h,v $
 * Revision 1.18  1997/12/10 17:29:37  ups
 * - Added a few prototypes needed for revised CHA and inlining
 *
 * Revision 1.17  1997/12/10 14:30:35  muller
 * - modification to the String Type
 * - cleanup in declarations
 *
 * Revision 1.16  1997/12/04 10:26:51  muller
 * - added new declarations
 *
 * Revision 1.15  1997/12/03 17:08:43  muller
 * - added declaration of exit_sthread
 *
 * Revision 1.14  1997/12/02 14:46:48  ups
 * - Changed CodeInfo so that it stores the corresponding DataState within
 *   the structure.  This was apperently necessary to get the inlining working
 *   without excessive memory usage.
 *
 * Revision 1.13  1997/11/28 11:12:18  ups
 * Changed uses of allocMemory to allocAtomicMemory where appropriate
 *
 * Revision 1.12  1997/11/26 08:29:09  muller
 * -field name update of the String structure (_F_)
 *
 * Revision 1.11  1997/11/21 15:52:18  ups
 * Imported seperate branch from ups: CC compability
 *
 * Revision 1.10  1997/11/17 17:17:07  muller
 * - proper init of interfaces and clinit method when downloading precompiled classes
 * - multithreaded environment for exceptions
 *
 * Revision 1.9  1997/10/27 10:04:16  ups
 * - Added (reverse) pointer from ClassMethod to MethodInfo to facilitate
 *   doing a call tree traversal while manipulating MethodInfo structures
 * - Moved data definitions from gen.h to this file, to support the pointer
 * - Added flag definition to support detection of thread subclasses
 *
 * Revision 1.8  1997/10/16 13:18:34  muller
 * - constant propagation
 * - variable propagation
 * - change in naming objects and classes for Tempo support
 *
 * Revision 1.7  1997/05/07 13:26:51  muller
 * - version publique alpha2.2
 *
 * Revision 1.6  1997/01/08 12:43:49  muller
 * -salsa alpha2 public release
 *
 * -redefining union for TEMPO is now supported globaly
 * -support for stubc implementation (fields moved in method definition)
 * -field added in ClassConstant to support TLONG proper reading
 *
 * Revision 1.5  1996/12/18 16:20:38  muller
 * - support for seperate VTable definition per classe
 * - support for guarding Monitors related calls to optimize single thread case
 *
 * Revision 1.4  1996/12/05 09:56:43  muller
 * specific struct for string constants
 *
 * Revision 1.3  1996/12/02 14:13:50  muller
 * Support for div by 0 and null pointers added
 *
 * Revision 1.2  1996/09/18 10:46:04  muller
 * babe -> salsa, preliminary TEMPO support
 *
 * Revision 1.1  1996/09/11 19:26:03  bellard
 * Initial revision
 *
 *
 */

#ifdef __GNUC__
#define COMPILER_ARRAY_MIN_SIZE 0
#else
/* Suif does not like array with a null length */
#define COMPILER_ARRAY_MIN_SIZE 1
#endif

/* an exception environment */
typedef struct _ExceptionEnv {
  struct _ExceptionEnv *ret_exception_env;  /* pointer to the previous
					       environment */
  jmp_buf jbuf;
} ExceptionEnv;

#include "my_thread.h"

/* header for any object */
typedef struct ObjectHeader {
  struct _VTable *vtable;     /* cache for the method table */
  int size;                   /* total size of the object. */
  struct monitor *mon;
} ObjectHeader;


typedef struct Object {
  ObjectHeader __header;
  char data[COMPILER_ARRAY_MIN_SIZE];
} Object;

#ifndef NULL
#define NULL ((void *)0)
#endif

/* an item stored on the stack of the Virtual Machine */
typedef union _StackItem {
  TREF r;
  TINT i;
  TLONG l;
  TFLOAT f;
  TDOUBLE d;
} StackItem;

#define GET_CLASS(r) (r->__header.vtable->class)

/* structure for a class */
typedef struct Class {
  ObjectHeader		__header;

  /* the following fields are extracted from the .class file */

  char *name;          /* name of the class */
  int access_flags;    /* access flags */
  
  char *super_name;    /* name of the superclass */

  union _ClassConstant	*constant_pool;   /* entries of the constant pool */

  char                  *constant_tags;   /* type of each entry of the 
					     constant pool */ 
  int                   constant_count;   /* number of entries */

  struct ClassMethod *methods;           /* methods of the class */
  int methods_count;                     /* # methods */

  struct _ClassField *fields;            /* idem fields */
  int fields_count;

  struct _ClassInterface *interfaces;    /* idem interfaces */
  int interfaces_count;
  
  /* the following fields are initialized when the class is resolved */

  struct Class *super;       /* superclass */

  int fields_offset;         /* offset of the end of dynamic fields (in bytes) */
  int fields_align_max;      /* alignement of the object structure (in bytes) */
  int instance_size;         /* size of an instance of the class (in bytes) */

  void *static_fields;     /* pointer to the static data */
  int static_fields_size;  /* size (in bytes) */

  struct _VTable *vtable;             /* virtual method table */
  int virtual_methods_count;          /* size of the vtable */
  
  struct Class *array_of_class;    /* If the class is an array of refs,
					 this is the class of the refs, 
					 NULL otherwise */
  char array_data_size;                /* size in bytes of each element of 
					  the array */
  char array_header_size;              /* the size of the header of the array */

  struct Class **ref_class;       /* array of classes referenced by the class,
                                    for precompiled libraries */
  
  int interface_id;                  /* the interface unique identifier */

  TREF class_loader;  /* class loader , NULL if none */

  struct Class *class_pool_next; /* next class in class_pool */

  struct Class *resolve_next;    /* next class to be resolved */
  int resolved;                  /* non nul if the class has been resolved */

  /* what are the subclasses of this class : these fields should be removed 
     from the class structure because they are only used in harissa2c */
  struct Class *sub_first;
  struct Class *sub_next;
} Class;

/* field structure */

typedef struct _ClassField {
  char *name;
  char *sig;
  int access_flags;
  
  /* the following fields are initialized when the class is resolved */
  int offset;

  /* the class to which the field belongs */
  Class *class;
  
  /* reference to a constant value in the constant pool */
  union _ClassConstant *value;

} ClassField;


/* method structure */

typedef struct MethodInfo _MethodInfo;

typedef struct ClassMethod {
  char *name;
  char *sig;
  int access_flags;

  /* the following fields are initialized when the class is resolved */

  /* offset in vtable */
  int offset;

  /* the class to which the method belongs */
  Class *class;

  /* the native code calls this method */
  void *native;

  /* the VM calls this function to execute the method */
  union _StackItem *(*native_stub)(struct ClassMethod *m,union _StackItem *vars); 
  
  /* the stub for calling a virtual method in bytecode from the C code */
  void *native_stubc;
  int args_size;  /* total number of arguments, in StackItem unit */
  int vars_size;  /* in StackItem unit */

  /* Non initialized fields in C code generated by hac */
  unsigned char *code;   /* vm code */
  int code_size;   /* in bytes */
  int stack_size; /* in StackItem unit */

  struct _MethodException *exceptions;
  int exceptions_count;

  struct BytecodeInfo *bytecode_info;

  /* This last field is needed by the global analysis */
  _MethodInfo *minfo;

      /* used when making list of methods */
  struct ClassMethod *next;
  
} ClassMethod;

/* Information related to the bytecode. TODO: most of the fields in ClassMethod
   should be put here. ClassMethod should only contain information independent
   of the implementation of the method */

typedef struct BytecodeInfo {

  /* line numbers */
  int line_number_table_length;
  struct LineNumberTable *line_number_table;

  /* local variable table */
  int local_variable_table_length;
  struct LocalVariableTable *local_variable_table;

} BytecodeInfo;

/* debug info */
typedef struct LineNumberTable {
  int start_pc;
  int line_number;
} LineNumberTable;

typedef struct LocalVariableTable {
  int start_pc;
  int end_pc;
  char *name;
  char *sig;
  int offset;
} LocalVariableTable;


/* exception structure */

typedef struct _MethodException {
  int start_pc;
  int end_pc;
  int handler_pc;
  int class_index;
} MethodException;

typedef struct _ClassInterface {
  char *name;       /* name of the interface */

  /* runtime data */
  Class *class;  /* corresponding class */
} ClassInterface;


/* table of methods. */
typedef struct _VTable { 
    struct Class *class;             /* the class of the object */     
    struct ClassMethod **vtable_vm;  /* the method table for the vm */
    void ***itable;                  /* the interface method table */ 
    void *native[1];                 /* the virtual method table */   
} VTable;

#define MK_VTable(t1)                                                  \
  typedef struct _VTable_ ## t1 {                                       \
    struct Class *class;             /* the class of the object */     \
    struct ClassMethod **vtable_vm;  /* the method table for the vm */ \
    void ***itable;                  /* the interface method table */  \
    void *native[t1];                /* the virtual method table */    \
  } VTable_ ## t1

#ifdef HARISSALIBS
  MK_VTable(COMPILER_ARRAY_MIN_SIZE);    /* Def for compiling librairies */
#endif
  
/*
 * arrays 
 */

typedef struct HArray {
  ObjectHeader __header;
  int length;
} HArray;

#define MK_ARRAY(t1,t2)				\
typedef struct ArrayOf ## t2 {			\
  ObjectHeader __header;			\
  int length;					\
  t1 data[COMPILER_ARRAY_MIN_SIZE];			\
} ArrayOf ## t2

 MK_ARRAY(TBYTE,Byte);
 MK_ARRAY(TSHORT,Short);
 MK_ARRAY(TCHAR,Char);
 MK_ARRAY(TINT,Integer);
 MK_ARRAY(TLONG,Long);
 MK_ARRAY(TFLOAT,Float);
 MK_ARRAY(TDOUBLE,Double);
 MK_ARRAY(TREF,Object);


/* TODO : make it simpler  */
#define TYPES_MAX 11

#define T_BOOL   4
#define T_CHAR   5
#define T_FLOAT  6
#define T_DOUBLE 7
#define T_BYTE   8
#define T_SHORT  9
#define T_INT    10
#define T_LONG   11
#define T_REF    12

#define T_VAR    0x10   /* Bit used to indicate that the cache contains a variable */
#define T_VAR_MASK 0x0F

typedef struct {
  char *name;
  Class *class;
} ArrayTypeInfo;

extern ArrayTypeInfo array_type_info[];
extern int type_size[];
extern int type_bsize[];
extern int type_balign[];

/* predefined basic classes */

#define CLASSNAME_OBJECT "java/lang/Object"
#define CLASSNAME_STRING "java/lang/String"
#define CLASSNAME_THREAD "java/lang/Thread"
#define CLASSNAME_CLASS  "java/lang/Class"

extern Class *ClassClass;
extern Class *ClassObject;
extern Class *ClassString;

/* access flags */

#define	ACC_PUBLIC		0x0001
#define	ACC_PRIVATE		0x0002
#define	ACC_PROTECTED		0x0004
#define	ACC_STATIC		0x0008
#define	ACC_FINAL		0x0010
#define	ACC_SYNCHRONIZED	0x0020
#define	ACC_VOLATILE		0x0040
#define	ACC_TRANSIENT		0x0080
#define	ACC_NATIVE		0x0100
#define	ACC_INTERFACE		0x0200
#define	ACC_ABSTRACT		0x0400

/* non standard flags. TODO: the compiled flag should be used when mixing
   native and interpreted classes */
#define ACC_COMPILED            0x0800
#define ACC_EXPLORED            0x1000
#define ACC_SUBCLASSOFTHREAD    0x2000

/*
 * Constant pool definitions
 */

#define	CONSTANT_Class			7
#define	CONSTANT_Fieldref		9
#define	CONSTANT_Methodref		10
#define	CONSTANT_InterfaceMethodref	11
#define	CONSTANT_String			8
#define	CONSTANT_Integer		3
#define	CONSTANT_Float			4
#define	CONSTANT_Long			5
#define	CONSTANT_Double			6
#define	CONSTANT_NameAndType		12
#define	CONSTANT_Utf8			1
#define	CONSTANT_Unicode		2

#define	CONSTANT_Unknown		0
#define	CONSTANT_Chararray		128

#define CONSTANT_String_res             127  /* a resolved string */
#define CONSTANT_Class_res              126  /* a resolved class */
#define CONSTANT_Methodref_res          125  /* a resolved method */

/* an entry of the constant pool */
typedef union _ClassConstant {
  TLONG c_long;
  struct {
    int low;
    int high;
  } mk_long;
  TDOUBLE c_double;
  TINT c_int;
  TFLOAT c_float;
  char *c_utf8;
  struct String *c_string;
  Class *c_class;
  ClassMethod *c_method;
  struct {
    int class_index;
    int nt_index;
  } c_ref;
  struct {
    int name_index;
    int sig_index;
  } c_nt;
} ClassConstant;

typedef struct _ClassConstantString {
  struct String *c_string;
  char dummy[sizeof(ClassConstant)-sizeof(struct String *)];
} ClassConstantString;

/* a java string */
typedef struct String {
  ObjectHeader __header;
  TREF _F_value;
  TINT _F_offset;
  TINT _F_count;
} String;

/* a java exception */
typedef struct Throwable {
 ObjectHeader __header;
 Object * backtrace;
 String * detailMessage;
} Throwable;

/* This is horrible. TODO: find a better solution */
#define Hjava_lang_Class  Class
#define Hjava_lang_Object Object
#define Hjava_lang_String String


/* Unicode.c */

int Utf8toUnicodeLength(char *src);
int Utf8toUnicode(TCHAR *dest,char *src);

int UnicodetoUtf8Length(TCHAR *src,int len);
int UnicodetoUtf8(char *dest,int dest_len,TCHAR *src,int len);

/* ClassHandler.c */

#define CLASSHASH_SIZE 512

extern int interface_id_count;

Class* lookupClass(Class *class_for_loader,char *c);
void resolveClass(Class *c);

void setClassInitFunc(void (*init_func)(ClassMethod *));
void setSystemNativeClasses(Class **resolve_tab,void **init_tab);

void initSystemClasses(void);

/* ClassLoader.c */

#define JAVAMAGIC       0xCafeBabe
#define MAJOR_VERSION   45
#define MINOR_VERSION   3

extern int load_debug_info;

Class *defineClass(unsigned char *class_buf, int class_size);
Class *defineClassArray(char *name);

/* list of external classes in library */
typedef struct ext_lib
{
  char *name;
  struct ext_lib *next;
}ext_lib;

extern ext_lib *ext_lib_name;
extern ext_lib symbol_head;


/* SystemClassLoader.c */

extern int Not_when_compiling;
void initSystemClassLoader(void);
Class *findSystemClass(char* cname);
void read_lib(char *current_lib);

#ifndef TEMPO
void errorClass(char *format,...);
#endif

/* Thread_support.c */
TREF harissa_currentThread(void);
void init_sthread();
void exit_sthread();
TREF HarissaThread[10];
int thread_count;

/* Memory.c */

void *allocMemory(int size);
void *allocAtomicMemory(int size);
void freeMemory(void *p);
TREF newObject(Class *class);
Class *newClass(void);
TREF newArray(Class *class,int length);
void *reallocMemory(char *p, int size);


/* Exception.c */

void initExceptions(void);

#ifndef __GNUC__
#define  __attribute__(x)
#endif
/* Attribute is supported only by GCC */
void throwException(TREF r) __attribute__ ((noreturn));
void throwArithmeticException()  __attribute__ ((noreturn));
void throwNullPointerException()  __attribute__ ((noreturn));
  
void Force_code();
  
extern TREF ClassFormatError;
extern TREF LinkageError;
extern TREF NoClassDefFoundError;
extern TREF NoSuchFieldError;
extern TREF NoSuchMethodError;
extern TREF OutOfMemoryError;
extern TREF UnsatisfiedLinkError;
extern TREF VirtualMachineError;
extern TREF ClassCircularityError;
extern TREF IncompatibleClassChangeError;

extern TREF NegativeArraySizeException;
extern TREF ClassCastException;
extern TREF IllegalMonitorStateException;
extern TREF NullPointerException;
extern TREF ArithmeticException;
extern TREF ArrayIndexOutOfBoundsException;
extern TREF ArrayStoreException;

/* Native.c */

int findNativeMethod(char *native_stub_name, ClassMethod *m);
void *findLibrarySymbol(char *symbol);
void initLibraryLoader(void);
void addNativeLib (char *lib);

/* vm.c */

#ifndef TEMPO
StackItem do_execute_java_method_vararg(void *env,TREF obj,
				    char *method_name,char *method_sig,
				    ClassMethod *m,int is_static,
				    va_list ap);
#endif

StackItem *execVM_Profile(ClassMethod *m,StackItem *vars);
StackItem *execVM_WithExceptions(ClassMethod *m,StackItem *vars);
StackItem *execVM_Synchronized(ClassMethod *m,StackItem *vars);
StackItem *execVM(ClassMethod *m,StackItem *vars);

extern int allow_bytecode_quick;
extern int use_profiler;
extern int init_system_classes_done;
void printProfile(void);
void initProfile(void);

/* ClassUtils.c */

int getSigTypeInt(char *sig);
int getSigType(char *sig);
char *getNextSig(char *sig);
int getArgsSize(char *sig);

int hasByteCode(ClassMethod *m);

Class *findClassOfArrayOfObject(Class *class,Class *c);
ClassMethod *findClassMethod(Class *c,char *name,char *sig);
ClassMethod *findMethod(Class *class,char *name,char *sig);

ClassConstant *findConstant1(Class *class,int c);
ClassConstant *findConstant(Class *class,int c,int type);
Class *findClassIndex(Class *class,int d);
ClassMethod *findClassMethodIndex(Class *class,int d);
ClassField *findClassFieldIndex(Class *class,int d);
ClassMethod *findInterfaceMethodIndex(Class *class,int d);
String *findClassStringIndex(Class *class,int d);

void genCIdent(char *str,char *ident);
void genMethodName(char *str,ClassMethod *m);

/* Threads_xx */
void MonitorEnterTh(TREF obj);
void MonitorExitTh(TREF obj);
void MonitorNotifyAll(TREF obj);
void MonitorNotify(TREF obj);
void MonitorWaitTh(TREF obj, TLONG millis);

/* misc */

int InstanceOf(Class *c,Class *class);
extern void (*_MonitorEnter)(TREF obj);
extern void (*_MonitorExit)(TREF obj);

int enterBabeVM(Class *native_class,
		void (*entry_point)(int,char **),
		int argc,char **argv,
		void (*exception_handler)(TREF obj));


#define IDENT_SIZE_MAX 512

Class *getFirstClass(void);
Class *getNextClass(void);

/*
 * These four datatype declarations had to be migrated from gen.h into
 * this file, since the global analysis needs a pointer from
 * ClassMethod to MethodInfo
 */

/* a structure giving additionnal information for TABLESWITCH & LOOKUPSWITCH */
typedef struct SwitchInfo {
  int low;
  int high;
  int keys_count;
  int keys[COMPILER_ARRAY_MIN_SIZE];
} SwitchInfo;

/*
 * This structure contains the data describing an IB instruction
 */ 
typedef  union {
    int i;
    TLONG l;
    TREF r;
    TFLOAT f;
    TDOUBLE d;
    ClassMethod *method;
    ClassField *field;
    Class *class;
    SwitchInfo *switch_info;
    char *utf8;
  } type_value;

/*
 * the abstract state of the VM at a given pc
 */
/* an abstract data value */

typedef struct DataValue {
  char type;     /* abstract type : T_INT, T_LONG, T_FLOAT, T_DOUBLE,
                    T_TREF, T_ADR, T_BAD */
  char is_const; /* true if constant */
  union {
    struct ClassList *class; /* the list of possible classes if T_REF type */
    int i;                   /* the constant value if T_INT and is_const == true */
  } v;
} DataValue;

/* 
 * For a reference, we store all the different classes possible. If it
 * is an array, we also store its abstract size, represented as an
 * abstract value of type int.  
 */

typedef struct ClassList {
  struct ClassList *next;  /* next possible class (NULL if none) */
  Class *class;            /* the class descriptor */
  DataValue array_size;    /* an integer */
} ClassList;

typedef struct DataState {
  int changed;         /* true if the DataState needs to be propagated */
  int sp;              /* SP value */
  int vars_size;       /* number of local variables (for memory
                          allocation only) */
  DataValue **stack;   /* abstract values for the stack */
  DataValue **vars;    /* abstract value for local variables */
} DataState;

/* ... */

typedef struct CodeInfo {

  int instr_size;            /* size of the instruction, in bytes */
  unsigned char instr_code;  /* IB opcode */

  char ins_count,outs_count; /* signature of the IB */ 
  char *ins;
  char *outs;

  int brs_count;             /* control flow information : addresses of 
				possible next instructions */
  int *brs;

  unsigned short var_offset; /* offset IB dealing with variables. It
                                can also contain a constant pool index
                                for LDC (string) (TODO: change it!)  */

  /* immediate value */
  type_value v;
  
  char has_label;             /* true if a label is before this 
				 IB instruction */
  char flag;                  /* this flag is used in genAnalyse1.c. 
				 (TODO: throw it away !) */
  int sp;                     /* static value of sp calculated by 
				 genAnalyse1.c */

  /* Associated DataState (not a pointer) */
  int dataStateInitialized;     /* Set to true when the structure actually contains data values */
  DataState dataState;
} CodeInfo;


/*
 * This structure contains all information caracterising an exception handler.
 */
typedef struct ExceptionInfo {
  struct ExceptionInfo *next;   /* pointer to the next exception 
				   handler (NULL if none) */
  int handler_pc;               /* pc of the handler */
  int start_pc;                 /* region where the handler is active:
				   start_pc <= pc < end_pc 
				   */
  int end_pc; 
  Class *class;                 /* the exception caught should be 
				   an instance of 'class' */
} ExceptionInfo;


/* 
 * This structure caracterises a method. It contains the an IB array
 * and all exception handlers. We need to duplicate the latter
 * information from 'method' because it may be changed by inlining.  
 */

typedef struct MethodInfo {
  ClassMethod *method;        /* method descriptor */
  CodeInfo *cinfo;            /* IB array */
  int code_size;              /* IB array size (in CodeInfo units) */
  int stack_size;             /* maximum stack size */
  int vars_size;              /* number of local variables */
  ExceptionInfo *einfo;       /* info for exceptions handlers */
  DataState **dataState;      /* Information computed by CHA, size is code_size */
  int flags;                  /* flags used by the analysis, see below */
} MethodInfo;

#define METHOD_INFO_REAL_CHA_DONE        1 /* Set if the CHA is usable for optimizations */
#define METHOD_INFO_THREAD_ANALYSIS_MARK 2 /* Used by the thread analysis to "color" the method */
#define METHOD_INFO_SPECIAL_CHA_DONE     4 /* Set if the CHA has been done but is unusable for optimizations */
